/*
    dbf.cpp - accesso in lettura ai file DBF
    Copyright (c) 1997 Simone Govoni
*/

#include <fstream.h>
#include <stdlib.h>
#include <string.h>
#pragma hdrstop

#include "types.h"
#include "dbf.h"


// Apri in lettura un file DBF
Dbf::Dbf(const char *name)
{
    Byte ft;  // terminatore di campo
    hd_allocated=FALSE;
    field_allocated=FALSE;
    record_allocated=FALSE;
    fdeleted=FALSE;
    curr_rec=0UL;
    is=new ifstream(name,ios::binary);
    if (is->good()) {
        // file aperto; leggi l'header
        hd=new Header;
        hd_allocated=TRUE;
        is->read((Byte *)hd,sizeof(Header));
        if (is->good()) {
            field_num=((hd->headerlen-1)-sizeof(Header))/(sizeof(Fielddescriptor));
            field=new Fielddescriptor[field_num];
            offset=new int[field_num];
            field_allocated=TRUE;
            // leggi i descrittori dei campi
            is->read((Byte *)field,sizeof(Fielddescriptor)*field_num);
            if (is->good()) {
                is->get(ft);    // leggi il terminatore di campo
                if (is->good() && ft==HEADER_TERMINATOR) {
                    compute_offset_table();
                    record=new char[hd->recordlen];
                    record_allocated=TRUE;
                    /*
                        La prossima seekg() e' richiesta poiche' il Clipper
                        inserisce uno 0 tra il terminatore dell'header e il
                        primo record. Questo byte extra rende i DBF prodotti
                        dal dBASE e quelli prodotti dal Clipper non
                        completamente equivalenti.
                    */
                    is->seekg(hd->headerlen);
                    skip();
                }
                else
                    error=BAD_HEADER;
            }
            else
                error=BAD_HEADER;
        }
        else
            error=BAD_HEADER;
    }
    else
        error=NOT_OPEN;
}


Dbf::~Dbf()
{
    if (field_allocated) {
        delete[] field;
        delete[] offset;
    }
    if (record_allocated)
        delete[] record;
    if (hd_allocated)
        delete hd;
    delete is;
}


// Calcola la tabella degli offset dei campi; L'offset di un campo e' la
// sua posizione all'interno di un record.
void Dbf::compute_offset_table(void)
{
    if (fcount()>0) {
        offset[0]=1;
        for (int i=1;i<fcount();i++)
            offset[i]=offset[i-1]+fieldsize(i);
    }
}


// Restituisce il nome del campo #fnum
void Dbf::fieldname(char *dst,int fnum)
{
    if (fieldnum_ok(fnum)) {
        strcpy(dst,(char *)field[fnum-1].name);
        error=OK;
    }
    else
        error=FIELD_NOT_EXIST;
}


// Restituisce il tipo del campo #fnum
Dbf::Typecode Dbf::fieldtype(int fnum)
{
    Typecode ftype;
    if (fieldnum_ok(fnum)) {
        switch (field[fnum-1].type) {
            case 'C':
                ftype=CHARACTER;
                break;
            case 'N':
                ftype=NUMERIC;
                break;
            case 'L':
                ftype=LOGICAL;
                break;
            case 'D':
                ftype=DATE;
                break;
            case 'M':
                ftype=MEMO;
                break;
            default:
                ftype=UNKNOWN;
                break;
        }
        error=OK;
    }
    else
        error=FIELD_NOT_EXIST;
    return ftype;
}


// Restituisce la dimensione del campo #fnum
unsigned int Dbf::fieldsize(int fnum)
{
    unsigned int flen;
    if (fieldnum_ok(fnum)) {
        flen=field[fnum-1].len;
        error=OK;
    }
    else
        error=FIELD_NOT_EXIST;
    return flen;
}


// Restituisce il contenuto del campo #fnum del record corrente
void Dbf::fieldget(char *dst,int fnum)
{
    if (fieldnum_ok(fnum)) {
        strncpy(dst,record+offset[fnum-1],field[fnum-1].len);
        dst[field[fnum-1].len]='\0';
        error=OK;
    }
    else
        error=FIELD_NOT_EXIST;
}


/*
    Restituisce TRUE se e' stata raggiunta la fine logica del DBF. La fine
    logica e' indicata dalla presenza del byte terminatore 0x1A.
*/
Bool Dbf::logicaleof()
{
    Bool logeof=FALSE;
    for (int i=0;i<=hd->recordlen-1 && record[i]!=END_OF_DBF;i++)
        ;
    if (record[i]==END_OF_DBF)
        logeof=TRUE;
    return logeof;
}


// Salta al record successivo
void Dbf::skip(void)
{
    Bool cont=TRUE;
    if (!is->eof()) {
        do {
            is->read((Byte *)record,hd->recordlen);
            if (is->eof()) {
                error=END_OF_FILE;
                cont=FALSE;
            }
            else if (is->good()) {
                curr_rec++;
                if (!logicaleof())
                    error=OK;
                else {
                    error=END_OF_FILE;
                    cont=FALSE;
                }
            }
            else {
                error=BAD_RECORD;
                cont=FALSE;
            }
        } while (fdeleted && deleted() && cont);
    }
    else
        error=END_OF_FILE;
}


// Salta direttamente al record #recn
void Dbf::go(unsigned long recn)
{
    if (recn>=1 && recn<=lastrec()) {
        is->seekg(hd->headerlen+(recn-1)*hd->recordlen);
        is->read((Byte *)record,hd->recordlen);
        curr_rec=recn;
        error=OK;
    }
    else
        error=INVALID_RECNO;
}


// Attiva/Disattiva il filtraggio dei record contrassegnati come cancellati
void Dbf::setdeleted(Bool del)
{
    fdeleted=del;
    is->seekg(hd->headerlen);   // salta al primo record
    curr_rec=0UL;
    skip();
}


// Ritorna TRUE se ftype e' un tipo "stampabile" direttamente
Bool Dbf::printable(Typecode ftype)
{
    return (ftype==CHARACTER || ftype==NUMERIC ||
           ftype==LOGICAL || ftype==DATE) ? TRUE : FALSE;
}
