/*
    dbf.h - definizione della classe Dbf per l'accesso ai file DBF
    Copyright (c) 1997 Simone Govoni
*/

/*
    Questo header deve essere incluso dopo i seguenti file:
    types.h, fstream.h
*/


#ifndef DBF_H
#define DBF_H

// Dbf: classe per l'accesso ai file DBF//////////////////////////////////

class Dbf {
public:
    // Codici d'errore
    enum Errorcode {OK=0,NOT_OPEN,BAD_HEADER,FIELD_NOT_EXIST,BAD_RECORD,
                    END_OF_FILE,INVALID_RECNO};
    // Tipi di campo
    enum Typecode {CHARACTER,NUMERIC,LOGICAL,DATE,MEMO,UNKNOWN};
    // Massima lunghezza del nome di un campo
    enum {MAX_FIELD_NAME=11};
    enum {MAX_FIELD_NUM=128};
private:
    // Header di un file DBF
    typedef struct {
        Byte version;   // versione
        Byte year;      // anno dell'ultimo aggiornamento
        Byte month;     // mese dell'ultimo aggiornamento
        Byte day;       // giorno dell'ultimo aggiornamento
        DWord reccount; // numero di record
        Word headerlen; // lunghezza dell'header
        Word recordlen; // lunghezza di un record
        Byte reserved[20];
    } Header;

    // Descrittore di campo
    typedef struct {
        Byte name[11];  // nome del campo
        Byte type;      // tipo del campo
        DWord data_adr; // indirizzo dei dati (data address)
        Byte len;       // lunghezza del campo
        Byte deccount;  // numero di decimali
        Byte reserved[14];
    } Fielddescriptor;

    enum {HEADER_TERMINATOR=0x0D};
    enum {END_OF_DBF=0x1A};
    enum {DELETED_TAG='*'};

    Header *hd;                 // header del file DBF
    Fielddescriptor *field;     // ptr. al primo descrittore di campo
    int field_num;              // numero di campi
    unsigned long curr_rec;     // numero del record corrente
    ifstream *is;               // input stream associato al file DBF
    Errorcode error;            // codice d'errore impostato dall'ultima operazione
    char *record;               // contenuto del record corrente
    int *offset;                // ptr. alla tabella degli offset dei campi
    Bool hd_allocated;          // TRUE se l'header e' stato allocato
    Bool field_allocated;       // TRUE se la tabella degli offset dei campi e' stata allocata
    Bool record_allocated;      // TRUE se il buffer del record e' stato allocato
    Bool fdeleted;              // TRUE se i record cancellati devono essere ignorati

    // Controlla la correttezza di un numero di campo
    Bool fieldnum_ok(int fnum) const
        { return (fnum >=1 && fnum<=field_num) ? TRUE : FALSE; }
    // Calcola la tabella degli offset dei campi
    void compute_offset_table(void);

    // Controlla se si e' raggiunta la fine logica del DBF
    Bool Dbf::logicaleof();
public:
    // Apri un file DBF in lettura
    Dbf(const char *name);

    ~Dbf();

    // Restituisce il numero di record nel file DBF
    unsigned long lastrec() const
        { return hd->reccount; }

    // Restituisce il numero del record corrente
    unsigned long recno() const
        { return curr_rec; }

    // Salta al record successivo
    void skip(void);

    // Restituisce il numero di campi del file DBF
    int fcount(void) const
        { return field_num; }

    // Restituisce il nome del campo #fnum
    void fieldname(char *dst,int fnum);

    // Restituisce il tipo del campo #fnum
    Typecode fieldtype(int fnum);

    // Restituisce lo stato d'errore dell'oggetto Dbf
    Errorcode geterror(void) const
        { return error; }

    // Attiva/Disattiva il filtraggio dei record cancellati
    void setdeleted(Bool del);

    // Restituisce lo stato del flag di cancellazione del record corrente
    Bool deleted(void) const
        { return record[0]==DELETED_TAG ? TRUE : FALSE; }

    // Restituisce la lunghezza del record
    unsigned int recsize(void) const
        { return hd->recordlen; }

    // Restituisce la lunghezza del campo #fnum
    unsigned int fieldsize(int fnum);

    // Controlla se il record pointer punta all'ultimo record
    Bool eof(void) const
        { return (is->eof() || error==END_OF_FILE ||
                  curr_rec==0 || curr_rec>lastrec()) ? TRUE : FALSE; }

    // Restituisce il contenuto del campo #fnum del record corrente
    void fieldget(char *dst,int fnum);

    // Salta direttamente al record #recn
    void go(unsigned long recn);

    // Ritorna TRUE se ftype e' un tipo stampabile direttamente
    Bool printable(Typecode ftype);

};

#endif // DBF_H