/*
    html.cpp - classe per la creazione di documenti HTML
    Copyright (c) 1997 Simone Govoni
*/

#include <fstream.h>
#include <stdlib.h>
#include <string.h>
#pragma hdrstop

#include "html.h"

/*
    Vettore per la traduzione dei caratteri ASCII, estesi e non, non
    immediatamente utilizzabili in HTML. La fonte delle entita' e' l'RFC-1866.
    NB: E' importante che il vettore sia ordinato in modo crescente in base al
    codice ASCII (il primo campo).
*/
Html::Entity Html::entities[]={
    {20,"&para;"},      // segno di paragrafo
    {21,"&sect;"},      // segno di sezione
    {34,"&quot;"},      // virgolette
    {38,"&amp;"},       // ampersand
    {60,"&lt;"},        // minore di
    {62,"&gt;"},        // maggiore di
    {128,"&Ccedil;"},   // "C" con cediglia
    {129,"&uuml;"},     // "u" con dieresi
    {130,"&eacute;"},   // "e" con accento acuto
    {131,"&acirc;"},    // "a" con accento circonflesso
    {132,"&auml;"},     // "a" con dieresi
    {133,"&agrave;"},   // "a" con accento grave
    {134,"&aring;"},    // "a" con cerchiolino
    {135,"&ccedil;"},   // "c" con cediglia
    {136,"&ecirc;"},    // "e" con accento circonflesso
    {137,"&euml;"},     // "e" con dieresi
    {138,"&egrave;"},   // "e" con accento grave
    {139,"&iuml;"},     // "i" con dieresi
    {140,"&icirc;"},    // "i" con accento circonflesso
    {141,"&igrave;"},   // "i" con accento grave
    {142,"&Auml;"},     // "A" con dieresi
    {143,"&Aring;"},    // "A" con cerchiolino
    {144,"&Eacute;"},   // "E" con accento acuto
    {145,"&aelig;"},    // dittongo "ae"
    {146,"&AElig;"},    // dittongo "AE"
    {147,"&ocirc;"},    // "o" con accento circonflesso
    {148,"&ouml;"},     // "o" con dieresi
    {149,"&ograve;"},   // "o" con accento grave
    {150,"&ucirc;"},    // "u" con accento circonflesso
    {151,"&ugrave;"},   // "u" con accento grave
    {152,"&yuml;"},     // "y" con dieresi
    {153,"&Ouml;"},     // "O" con dieresi
    {154,"&Uuml;"},     // "U" con dieresi
    {155,"&cent;"},     // centesimo
    {156,"&pound;"},    // sterlina
    {157,"&yen;"},      // yen
    {160,"&aacute;"},   // "a" con accento acuto
    {161,"&iacute;"},   // "i" con accento acuto
    {162,"&oacute;"},   // "o" con accento acuto
    {163,"&uacute;"},   // "u" con accento acuto
    {164,"&ntilde;"},   // "n" con tilde
    {165,"&Ntilde;"},   // "N" con tilde
    {166,"&ordf;"},     // indicatore ordinale femminile
    {167,"&ordm;"},     // indicatore ordinale maschile
    {168,"&iquest;"},   // punto interrogativo rovesciato
    {170,"&not;"},      // negazione
    {171,"&frac12;"},   // 1/2
    {172,"&frac14;"},   // 1/4
    {173,"&iexcl;"},    // punto esclamativo rovesciato
    {174,"&laquo;"},    // virgolette angolate a sinistra
    {175,"&raquo;"},    // virgolette angolate a destra
    {230,"&micro;"},    // micro (mu)
    {241,"&plusmn;"},   // piu' o meno
    {248,"&deg;"},      // gradi
    {253,"&sup2;"}      // "2" all'apice
};


// Apri in scrittura un file HTML
Html::Html(const char *filename)
    : entitynum(sizeof(entities)/sizeof(Entity))
{
    ofs=new ofstream(filename);
    if (ofs->good()) {
        /*
            L'RFC-1866 "Hypertext Markup Language - 2.0" afferma:

            "To identify information as an HTML document conforming to this
            specification, each document must start with one of the following
            document type declarations.

            <!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">
            .
            . snipped...
            .
            NOTE - If the body of a `text/html' message entity does not begin
            with a document type declaration, an HTML user agent should infer
            the above document type declaration.

            <!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0 Level 2//EN">
            .
            . snipped...
            .
            <!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0 Level 1//EN">
            .
            . snipped...
            .
            <!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0 Strict//EN">
            .
            . snipped
            .
            <!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0 Strict Level 1//EN">"

            Per cui mi adeguo. Nella pratica tutti i browser provati fanno
            a meno del <DOCTYPE ...>.
        */

        *ofs << "<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">\n"
                "<HTML>\n";

        error=Html::OK;
    }
    else
        error=Html::NOT_OPEN;
}


Html::~Html()
{
    if (ofs->good())
        *ofs << "</HTML>\n";
    delete ofs;
}


/*
    Cerca l'entita' corrispondente al carattere ascii c; se non ne trova una
    restituisce NULL.
    L'algoritmo di ricerca binaria e' una leggera variazione di quello
    riportato del libro "Algoritmi in C++" di R.Sedgewick, pag.205,
    Ed.Italiana.
*/
char *Html::searchentity(unsigned char c)
{
    int l=0;                // estremo sinistro dell'area di ricerca
    int r=entitynum-1;      // estremo destro dell'area di ricerca
    int x;                  // centro dell'area di ricerca
    while (r>=l) {
        x=(l+r)/2;
        if (c==entities[x].xascii)
            r=l-1;      // forza l'uscita dal ciclo while
        else if (c<entities[x].xascii)
            r=x-1;      // continua la ricerca alla sinistra di x
        else
            l=x+1;      // continua la ricerca alla destra di x
    }
    return (c==entities[x].xascii ? entities[x].htmlcode : NULL);
}


/*
    Invia sullo stream di output una stringa convertendo in entita' i
    caratteri che in HTML hanno una forma particolare. I caratteri che non
    hanno equivalente in HTML sono ricopiati senza variazioni.
*/
void Html::puttext(const char *s)
{
    int len=strlen(s);
    char c;
    char *ent;      // entita' corrispondente al carattere scandito
    for (int i=0;i<len;i++)
        if ((ent=searchentity(c=s[i]))!=NULL)
            *ofs << ent;
        else
            ofs->put(c);
}


// Inserisce un commento nel documento HTML
void Html::comment(const char *s)
{
    *ofs << "<!-- " << s << " -->\n";
}


void Html::openhead(void)
{
    *ofs << "<HEAD>\n";
}


void Html::closehead(void)
{
    *ofs << "</HEAD>\n";
}


void Html::title(const char *s)
{
    *ofs << "<TITLE>";
    puttext(s);
    *ofs << "</TITLE>\n";
}


void Html::openbody(void)
{
    *ofs << "<BODY>\n";
}


void Html::closebody(void)
{
    *ofs << "</BODY>\n";
}


void Html::heading(int level,const char *s)
{
    if (level>=1 && level<=6) {
        char buf[2];
        itoa(level,buf,10);
        *ofs << "<H" << buf << '>';
        puttext(s);
        *ofs << "</H" << buf << ">\n";
    }
}


void Html::openpar(void)
{
    *ofs << "<P>\n";
}


void Html::closepar(void)
{
    *ofs << "</P>\n";
}


void Html::linebreak(void)
{
    *ofs << "<BR>\n";
}


void Html::opencenter(void)
{
    *ofs << "<CENTER>";
}


void Html::closecenter(void)
{
    *ofs << "</CENTER>";
}


void Html::hrule(void)
{
    *ofs << "<HR>\n";
}


void Html::hrule(int size)
{
    char buf[10];
    itoa(size,buf,10);
    *ofs << "<HR SIZE=" << buf << ">\n";
}


void Html::opentable(int border)
{
    if (border)
        *ofs << "<TABLE BORDER=" << border << ">\n";
    else
        *ofs << "<TABLE>\n";
}


void Html::closetable(void)
{
    *ofs << "</TABLE>\n";
}


void Html::openrecord(void)
{
    *ofs << "<TR>\n";
}


void Html::closerecord(void)
{
    *ofs << "</TR>\n";
}


void Html::fielddescript(const char *s)
{
    *ofs << "<TH>";
    puttext(s);
    *ofs << "</TH>\n";
}


void Html::fieldcontent(const char *s,Html::Alignmode amode)
{
    *ofs << "<TD";
    switch (amode) {
        case LEFT:
            *ofs << '>';
            break;
        case CENTER:
            *ofs << " ALIGN=CENTER>";
            break;
        case RIGHT:
            *ofs << " ALIGN=RIGHT>";
            break;
        default:
            *ofs << '>';
            break;
    }
    puttext(s);
    *ofs << "</TD>\n";
}


void Html::caption(char *s,Html::Alignmode alcapt)
{
    *ofs << "<CAPTION";
    switch (alcapt) {
        case TOP:
            *ofs << " ALIGN=TOP>";
            break;
        case BOTTOM:
            *ofs << " ALIGN=BOTTOM>";
            break;
        default:
            *ofs << '>';
            break;
    }
    puttext(s);
    *ofs << "</CAPTION>\n";
}
